<?php

namespace Drupal\devutils;

use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Config\ConfigException;
use Drupal\Core\Config\FileStorage;
use Drupal\Core\Config\InstallStorage;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\Extension\ThemeExtensionList;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\language\Config\LanguageConfigFactoryOverride;

/**
 * Update configs from yml files.
 */
class UpdateConfigs implements UpdateConfigsInterface {

  /**
   * The module extension list.
   *
   * @var \Drupal\Core\Extension\ModuleExtensionList
   */
  protected $extensionListModule;

  /**
   * The theme extension list.
   *
   * @var \Drupal\Core\Extension\ThemeExtensionList
   */
  protected $themeList;

  /**
   * The active configuration storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $activeStorage;

  /**
   * The logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs an UpdateConfigs object.
   *
   * @param \Drupal\Core\Extension\ModuleExtensionList $extension_list_module
   *   The module extension list.
   * @param \Drupal\Core\Extension\ThemeExtensionList $theme_list
   *   The theme extension list.
   * @param \Drupal\Core\Config\StorageInterface $active_storage
   *   The active configuration storage.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The active configuration storage.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   */
  public function __construct(ModuleExtensionList $extension_list_module,
                              ThemeExtensionList $theme_list,
                              StorageInterface $active_storage,
                              LoggerChannelFactoryInterface $logger_factory,
                              EntityTypeManagerInterface $entity_type_manager) {
    $this->extensionListModule = $extension_list_module;
    $this->themeList = $theme_list;
    $this->activeStorage = $active_storage;
    $this->logger = $logger_factory->get('devutils');
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function module($name, array $configs) {
    try {
      $path = $this->extensionListModule->getPath($name);
      $this->configsUpdate($name, $configs, $path);
    }
    catch (\Exception $e) {
      $this->logger->error($e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function modules(array $configs) {
    foreach ($configs as $name => $config) {
      $this->module($name, $config);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function theme(string $theme, array $configs) {
    try {
      $path = $this->themeList->getPath($theme);
      $this->configsUpdate($theme, $configs, $path);
    }
    catch (\Exception $e) {
      $this->logger->error($e->getMessage());
    }
  }

  /**
   * Update the language collection.
   *
   * @param \Drupal\Core\Config\FileStorage $source
   *   The FileStorage source.
   * @param string $path
   *   The config extension (module, profile) path.
   * @param string $config
   *   The config names.
   * @param string $directory
   *   The config directory.
   */
  private function languageCollectionUpdate(FileStorage $source, string $path, string $config, string $directory) {
    foreach ($source->getAllCollectionNames() as $collection) {
      if (!str_contains($collection, 'language.')) {
        continue;
      }

      $sourceColl = new FileStorage($path . '/' . $directory, $collection);
      if (!$sourceColl->exists($config)) {
        continue;
      }

      if (!\Drupal::hasService('language.config_factory_override')) {
        return;
      }
      /** @var \Drupal\language\Config\LanguageConfigFactoryOverride $configFactoryOverride */
      $configFactoryOverride = \Drupal::service('language.config_factory_override');

      $language = str_replace('language.', '', $collection);
      $data = $sourceColl->read($config);
      if (!$configFactoryOverride->getStorage($language)
        ->write($config, $data)) {
        throw new ConfigException("The configuration ${config} not saved in collection ${collection}, something was wrong.");
      }

    }
  }

  /**
   * Update configs in path.
   *
   * @param string $name
   *   The machine name of the module or theme.
   * @param array $configs
   *   Configs to update.
   * @param string $path
   *   Directory where are the configs.
   */
  private function configsUpdate(string $name, array $configs, string $path) {
    $configInstallDirectory = new FileStorage($path . '/' . InstallStorage::CONFIG_INSTALL_DIRECTORY);
    $configOptionalDirectory = new FileStorage($path . '/' . InstallStorage::CONFIG_OPTIONAL_DIRECTORY);

    foreach ($configs as $config) {
      $ymlData = NULL;
      if ($configInstallDirectory->exists($config)) {
        $ymlData = $configInstallDirectory->read($config);
      }
      elseif ($configOptionalDirectory->exists($config)) {
        $ymlData = $configOptionalDirectory->read($config);
      }

      if (!$ymlData) {
        throw new ConfigException("The configuration ${config} not found in extension ${name}");
      }

      if (str_contains($config, '.storage.')) {
        $this->createStorage($ymlData);
        continue;
      }

      if (!$this->activeStorage->write($config, $ymlData)) {
        throw new ConfigException("Configuration ${config} not saved in module ${name}, something wrong happened.");
      }

      $this->languageCollectionUpdate($configInstallDirectory, $path, $config, InstallStorage::CONFIG_INSTALL_DIRECTORY);
    }
  }

  /**
   * Create a storage from config data.
   *
   * @param array $data
   *   The storage data.
   */
  private function createStorage(array $data) {
    try {
      if (!isset($data['entity_type']) || !isset($data['field_name'])) {
        $this->logger->error("The storage data not contains entity_type or field_name key.");
        return;
      }

      /** @var \Drupal\Core\Config\Entity\ConfigEntityStorageInterface $fieldStorageConfig */
      $fieldStorageConfig = $this->entityTypeManager->getStorage('field_storage_config');
      $fieldId = $data['entity_type'] . '.' . $data['field_name'];
      if ($fieldStorageConfig->load($fieldId)) {
        $this->logger->warning("The field storage <${$fieldId}> already exists.");
        return;
      }

      $fieldStorageConfig->createFromStorageRecord($data)->save();
    }
    catch (EntityStorageException | InvalidPluginDefinitionException | PluginNotFoundException $e) {
      $this->logger->error($e->getMessage());
    }
  }

}
